{  Autor: Rafael Alberto Moreno Parra
  Sitio Web:  http://darwin.50webs.com
  Correo:  enginelife@hotmail.com

  Evaluador de expresiones algebraicas por ejemplo, 57+1.87/84.89-(6.8*e+b+8769-4*b/8+b^2*4^e/f)+5.4-(d/9.6+0.2) con las siguientes funcionalidades:
  1. Ilimitado nmero de parntesis
  2. Ms rpido y ms sencillo que el evaluador escrito para el primer libro: "Desarrollo de un evaluador de expresiones algebraicas"
  3. Manejo de 26 variables
  4. Manejo de 12 funciones
  5. Manejo de operadores +, -, *, /, ^ (potencia)
  6. Manejo del menos unario: lo reemplaza con (0-1)# donde # es la operacin con mayor relevancia y equivale a la multiplicacin

  Versin: 2.00 [Genera el libro "Desarrollo de un evaluador de expresiones algebraicas. Versin 2.0"]
  Fecha: Enero de 2013
  Licencia: LGPL

  Algoritmo:

  Se toma una expresin como 7*x+sen(12.8/y+9)-5*cos(9-(8.3/5.11^3-4.7)*7.12)+0.445
  Se agregan parntesis de inicio y fin  (7*x+sen(12.8/y+9)-5*cos(9-(8.3/5.11^3-4.7)*7.12)+0.445)
  Se divide en piezas simples   | ( | 7 | * | x | + | sen( | 12.8 | / | y | + | 9 | ) | - | 5 | * | cos( | 9 | - | ( | 8.3 | / | 5.11 | ^ | 3 | - | 4.7 | ) | * | 7.12 | ) | + | 0.445 | ) |
  Esas piezas estn clasificadas:
   Parntesis que abre (
   Parntesis que cierra )
   Nmeros  7   12.8  9   5
   Variables  x  y
   Operadores + - * / ^
   Funciones  sen(  cos(
  Luego se convierte esa expresin larga en expresiones cortas de ejecucin del tipo
  Acumula = operando(nmero/variable/acumula)  operador(+, -, *, /, ^)   operando(nmero/variable/acumula)
    [0]  5.11 ^ 3
    [1]  8.3  / [0]
    [2]  [1] - 4.7
    [3]  [2] + 0
    [4]  [3] * 7,12
    [5]  9   - [4]
    [6]  cos([5])
    [7]  12,8 / y
    [8]  [7] +  9
    [9]  sen([8])
    [10] 7 * x
    [11] 5 * [6]
    [12] [10] + [9]
    [13] [12] - [11]
    [14] [13] + 0.445
    [15] [14] + 0
  La expresin ya est analizada y lista para evaluar.
  Se evala yendo de [0] a [15], en [15] est el valor final.
}
unit Evaluar;

interface
uses
  //Requerido para el TobjectList
  Contnrs, SysUtils, Math, Pieza_Simple, Pieza_Ejecuta;
type
  TEvaluar = class
  private
    { Esta constante sirve para que se reste al carcter y se obtenga el nmero.  Ejemplo:  '7' - ASCIINUMERO =  7 }
    ASCIINUMERO: integer;

    { Esta constante sirve para que se reste al carcter y se obtenga el nmero de la letra. Ejemplo:  'b' - ASCIILETRA =  1 }
    ASCIILETRA: integer;

    { Las funciones que soporta este evaluador }
    TAMANOFUNCION: integer;
    listaFunciones: string;

    { Constantes de los diferentes tipos de datos que tendrn las piezas }
    ESFUNCION: integer;
    ESPARABRE: integer;
    ESPARCIERRA: integer;
    ESOPERADOR: integer;
    ESNUMERO: integer;
    ESVARIABLE: integer;

    //Listado de Piezas de anlisis
    PiezaSimple: TobjectList;
    objPiezaSimple: TPieza_Simple;

    //Listado de Piezas de ejecucin
    PiezaEjecuta: TobjectList;
    objPiezaEjecuta: TPieza_Ejecuta;

    Contador_Acumula: integer;

    //Almacena los valores de las 26 diferentes variables que puede tener la expresin algebraica
    VariableAlgebra: array[0..26] of double;

    function DobleTripleOperadorSeguido(expr: string): boolean;
    function OperadorParentesisCierra(expr: string): boolean;
    function ParentesisAbreOperador(expr: string): boolean;
    function ParentesisDesbalanceados(expr: string): boolean;
    function ParentesisVacio(expr: string): boolean;
    function ParentesisBalanceIncorrecto(expr: string): boolean;
    function ParentesisCierraNumero(expr: string): boolean;
    function NumeroParentesisAbre(expr: string): boolean;
    function DoblePuntoNumero(expr: string): boolean;
    function ParentesisCierraVariable(expr: string): boolean;
    function VariableluegoPunto(expr: string): boolean;
    function PuntoluegoVariable(expr: string): boolean;
    function NumeroAntesVariable(expr: string): boolean;
    function VariableDespuesNumero(expr: string): boolean;
    function Chequea4letras(expr: string): boolean;
    function FuncionInvalida(expr: string): boolean;
    function EsFuncionInvalida(car1: char; car2: char; car3: char): boolean;
    function VariableInvalida(expr: string): boolean;
    function VariableParentesisAbre(expr: string): boolean;
    function ParCierraParAbre(expr: string): boolean;
    function OperadorPunto(expr: string): boolean;
    function ParAbrePunto(expr: string): boolean;
    function PuntoParAbre(expr: string): boolean;
    function ParCierraPunto(expr: string): boolean;
    function PuntoOperador(expr: string): boolean;
    function PuntoParCierra(expr: string): boolean;

    procedure Generar_Piezas_Simples(expresion: string);
    procedure Generar_Piezas_Ejecucion();
    procedure Generar_Piezas_Operador(operA: char; operB: char; inicio: integer);
  public
    Constructor Create();
    function EvaluaSintaxis(expresion: string): integer;
    function MensajeSintaxis(CodigoError: integer): string;
    function ArreglaNegativos(expresion: string): string;
    function TransformaExpresion(expr: string): string;
    procedure Analizar(expresion: string);
    function Calcular(): double;
    procedure ValorVariable(variableAlgebra: char; valor: double);
  end;

implementation

constructor TEvaluar.Create();
begin
   //Constantes de los diferentes tipos de datos que tendrn las piezas */
   self.ESFUNCION := 1;
   self.ESPARABRE := 2;
   self.ESPARCIERRA := 3;
   self.ESOPERADOR := 4;
   self.ESNUMERO := 5;
   self.ESVARIABLE := 6;
   self.TAMANOFUNCION := 39;
   self.ASCIINUMERO := 48;
   self.ASCIILETRA := 97;
   self.listaFunciones := 'sinsencostanabsasnacsatnlogceiexpsqrrcb';
end;

//Valida la expresin algebraica
function TEvaluar.EvaluaSintaxis(expresion: string): integer;
begin
  //Hace 25 pruebas de sintaxis
  if DobleTripleOperadorSeguido(expresion)then begin Result :=  1; exit; end;
  if OperadorParentesisCierra(expresion)then begin Result :=  2; exit; end;
  if ParentesisAbreOperador(expresion)then begin Result :=  3; exit; end;
  if ParentesisDesbalanceados(expresion)then begin Result :=  4; exit; end;
  if ParentesisVacio(expresion)then begin Result :=  5; exit; end;
  if ParentesisBalanceIncorrecto(expresion)then begin Result :=  6; exit; end;
  if ParentesisCierraNumero(expresion)then begin Result :=  7; exit; end;
  if NumeroParentesisAbre(expresion)then begin Result :=  8; exit; end;
  if DoblePuntoNumero(expresion)then begin Result :=  9; exit; end;
  if ParentesisCierraVariable(expresion)then begin Result :=  10; exit; end;
  if VariableluegoPunto(expresion)then begin Result :=  11; exit; end;
  if PuntoluegoVariable(expresion)then begin Result :=  12; exit; end;
  if NumeroAntesVariable(expresion)then begin Result :=  13; exit; end;
  if VariableDespuesNumero(expresion)then begin Result :=  14; exit; end;
  if Chequea4letras(expresion)then begin Result :=  15; exit; end;
  if FuncionInvalida(expresion)then begin Result :=  16; exit; end;
  if VariableInvalida(expresion)then begin Result :=  17; exit; end;
  if VariableParentesisAbre(expresion)then begin Result :=  18; exit; end;
  if ParCierraParAbre(expresion)then begin Result :=  19; exit; end;
  if OperadorPunto(expresion)then begin Result :=  20; exit; end;
  if ParAbrePunto(expresion)then begin Result :=  21; exit; end;
  if PuntoParAbre(expresion)then begin Result :=  22; exit; end;
  if ParCierraPunto(expresion)then begin Result :=  23; exit; end;
  if PuntoOperador(expresion)then begin Result :=  24; exit; end;
  if PuntoParCierra(expresion)then begin Result :=  25; exit; end;
  Result := 0; //No se detect error de sintaxis
end;

//Muestra mensaje de error sintctico
function TEvaluar.MensajeSintaxis(CodigoError: integer): string;
begin
  case CodigoError of
      0: begin  Result := 'No se detect error sintctico en las 25 pruebas que se hicieron.'; exit; end;
      1: begin  Result := '1. Dos o ms operadores estn seguidos. Ejemplo: begin 2++4, 5-*3'; exit; end;
      2: begin  Result := '2. Un operador seguido de un parntesis que cierra. Ejemplo: begin 2-(4+)-7'; exit; end;
      3: begin  Result := '3. Un parntesis que abre seguido de un operador. Ejemplo: begin 2-(*3)'; exit; end;
      4: begin  Result := '4. Que los parntesis estn desbalanceados. Ejemplo: begin 3-(2*4))'; exit; end;
      5: begin  Result := '5. Que haya parntesis vaco. Ejemplo: begin 2-()*3'; exit; end;
      6: begin  Result := '6. Parntesis que abre no corresponde con el que cierra. Ejemplo: begin 2+3)-2*(4'; exit; end;
      7: begin Result := '7. Un parntesis que cierra y sigue un nmero. Ejemplo: begin (3-5)7-(1+2)'; exit; end;
      8: begin Result := '8. Un nmero seguido de un parntesis que abre. Ejemplo: begin 7-2(5-6)'; exit; end;
      9: begin Result := '9. Doble punto en un nmero de tipo real. Ejemplo: begin 3-2..4+1  7-6.46.1+2'; exit; end;
      10: begin Result := '10. Un parntesis que cierra seguido de una variable. Ejemplo: begin (12-4)y-1'; exit; end;
      11: begin Result := '11. Una variable seguida de un punto. Ejemplo: begin 4-z.1+3'; exit; end;
      12: begin Result := '12. Un punto seguido de una variable. Ejemplo: begin 7-2.p+1'; exit; end;
      13: begin Result := '13. Un nmero antes de una variable. Ejemplo: begin 3x+1'; exit; end;
      14: begin Result := '14. Un nmero despus de una variable. Ejemplo: begin x21+4'; exit; end;
      15: begin Result := '15. Hay 4 o ms letras seguidas. Ejemplo: begin 12+ramp+8.9'; exit; end;
      16: begin Result := '16. Funcin inexistente. Ejemplo: begin 5*alo(78)'; exit; end;
      17: begin Result := '17. Variable invlida (solo pueden tener una letra). Ejemplo: begin 5+tr-xc+5'; exit; end;
      18: begin Result := '18. Variable seguida de parntesis que abre. Ejemplo: begin 5-a(7+3)'; exit; end;
      19: begin Result := '19. Despus de parntesis que cierra sigue parntesis que abre. Ejemplo: begin (4-5)(2*x)'; exit; end;
      20: begin Result := '20. Despus de operador sigue un punto. Ejemplo: begin -.3+7'; exit; end;
      21: begin Result := '21. Despus de parntesis que abre sigue un punto. Ejemplo: begin 3*(.5+4)'; exit; end;
      22: begin Result := '22. Un punto seguido de un parntesis que abre. Ejemplo: begin 7+3.(2+6)'; exit; end;
      23: begin Result := '23. Parntesis cierra y sigue punto. Ejemplo: begin (4+5).7-2'; exit; end;
      24: begin Result := '24. Punto seguido de operador. Ejemplo: begin 5.*9+1'; exit; end;
      else begin Result := '25. Punto seguido de parntesis que cierra. Ejemplo: begin (3+2.)*5'; exit; end;
  end;
end;

//Retira los espacios y caracteres invlidos.
function TEvaluar.TransformaExpresion(expr: string): string;
var
  validos, nuevaExpr, expr2: string;
  pos, valida: integer;
  letra: char;
begin
  validos := 'abcdefghijklmnopqrstuvwxyz0123456789.+-*/^()';
  expr2 := lowercase(expr);
  nuevaExpr := '(';
  for pos := 1 to length(expr2) do
  begin
    letra := expr2[pos];
    for valida := 1 to length(validos) do
    begin
      if letra = validos[valida] then
      begin
          nuevaExpr := nuevaExpr + letra;
          break;
      end;
    end;
  end;
  nuevaExpr := nuevaExpr + ')';
  Result := nuevaExpr;
end;

//1. Dos o ms operadores estn seguidos. Ejemplo: 2++4, 5-*3
function TEvaluar.DobleTripleOperadorSeguido(expr: string): boolean;
var
  pos: integer;
  car1, car2, car3: char;
begin
  for pos := 1 to length(expr) do
  begin
    car1 := expr[pos];   //Extrae un carcter
    car2 := expr[pos + 1]; //Extrae el siguiente carcter

    //Compara si el carcter y el siguiente son operadores, dado el caso retorna true
    if (car1 = '+') or (car1 = '-') or (car1 = '*') or (car1 = '/') or (car1 = '^') then
      if (car2 = '+') or (car2 = '*') or (car2 = '/') or (car2 = '^') then
      begin
        Result := true;
        exit;
      end;
  end;

  for pos := 1 to length(expr) - 1 do
  begin
    car1 := expr[pos];   //Extrae un carcter
    car2 := expr[pos + 1]; //Extrae el siguiente carcter
    car3 := expr[pos + 2]; //Extrae el siguiente carcter

    //Compara si el carcter y el siguiente son operadores, dado el caso retorna true
    if (car1 = '+') or (car1 = '-') or (car1 = '*') or (car1 = '/') or (car1 = '^') then
      if (car2 = '+') or (car2 = '-') or (car2 = '*') or (car2 = '/') or (car2 = '^') then
        if (car3 = '+') or (car3 = '-') or (car3 = '*') or (car3 = '/') or (car3 = '^') then
        begin
          Result := true;
          exit;
        end;
  end;

  Result := false;  //No encontr doble/triple operador seguido
end;

//2. Un operador seguido de un parntesis que cierra. Ejemplo: 2-(4+)-7
function TEvaluar.OperadorParentesisCierra(expr: string): boolean;
var
  pos: integer;
  car1: char;
begin
  for pos := 1 to length(expr) do
  begin
    car1 := expr[pos];   //Extrae un carcter

    //Compara si el primer carcter es operador y el siguiente es parntesis que cierra
    if (car1 = '+') or (car1 = '-') or (car1 = '*') or (car1 = '/') or (car1 = '^') then
      if (expr[pos + 1] = ')') then
        begin
          Result := true;
          exit;
        end;
  end;
  Result := false; //No encontr operador seguido de un parntesis que cierra
end;

//3. Un parntesis que abre seguido de un operador. Ejemplo: 2-(*3)
function TEvaluar.ParentesisAbreOperador(expr: string): boolean;
var
  pos: integer;
  car2: char;
begin
  for pos := 1 to length(expr) do
  begin
    car2 := expr[pos + 1]; //Extrae el siguiente carcter

    //Compara si el primer carcter es parntesis que abre y el siguiente es operador
    if expr[pos] = '(' then
      if (car2 = '+') or (car2 = '*') or (car2 = '/') or (car2 = '^') then
      begin
        Result := true;
        Exit;
      end;
  end;
  Result := false;  //No encontr parntesis que abre seguido de un operador
end;

//4. Que los parntesis estn desbalanceados. Ejemplo: 3-(2*4))
function TEvaluar.ParentesisDesbalanceados(expr: string): boolean;
var
  parabre, parcierra, pos: integer;
  car1: char;
begin
  parabre := 0; parcierra := 0;
  for pos := 1 to length(expr) do
  begin
    car1 := expr[pos];
    if car1 = '(' then Inc(parabre);
    if car1 = ')' then Inc(parcierra);
  end;
  if parabre <> parcierra then
    Result := true
  else
    Result := false;
end;

//5. Que haya parntesis vaco. Ejemplo: 2-()*3
function TEvaluar.ParentesisVacio(expr: string): boolean;
var
  pos: integer;
begin
  //Compara si el primer carcter es parntesis que abre y el siguiente es parntesis que cierra
  for pos := 1 to length(expr) do
  begin
    if (expr[pos] = '(') and (expr[pos + 1] = ')') then
    begin
      Result := true;
      Exit;
    end;
  end;
  Result := false;
end;

//6. As estn balanceados los parntesis no corresponde el que abre con el que cierra. Ejemplo: 2+3)-2*(4
function TEvaluar.ParentesisBalanceIncorrecto(expr: string): boolean;
var
  balance, pos: integer;
  car1: char;
begin
  balance := 0;
  for pos := 1 to length(expr) do
  begin
    car1 := expr[pos];   //Extrae un carcter
    if car1 = '(' then Inc(balance);
    if car1 = ')' then Dec(balance);
    if balance < 0 then begin Result := true; Exit; end;
  end;
  Result := false;
end;

//7. Un parntesis que cierra y sigue un nmero o parntesis que abre. Ejemplo: (3-5)7-(1+2)(3/6)
function TEvaluar.ParentesisCierraNumero(expr: string): boolean;
var
  pos: integer;
  car2: char;
begin
  for pos := 1 to length(expr) do
  begin
    car2 := expr[pos + 1]; //Extrae el siguiente carcter

    //Compara si el primer carcter es parntesis que cierra y el siguiente es nmero
    if (expr[pos] = ')') then
      if (car2 >= '0') and (car2 <= '9')  then
      begin
        Result := true;
        Exit;
      end;
  end;
  Result := false;
end;


//8. Un nmero seguido de un parntesis que abre. Ejemplo: 7-2(5-6)
function TEvaluar.NumeroParentesisAbre(expr: string): boolean;
var
  pos: integer;
  car1: char;
begin
  for pos := 1 to length(expr) do
  begin
    car1 := expr[pos];   //Extrae un carcter

    //Compara si el primer carcter es nmero y el siguiente es parntesis que abre
    if (car1 >= '0') and (car1 <= '9') then
      if expr[pos + 1] = '(' then
      begin
        Result := true;
        Exit;
      end;
  end;
  Result := false;
end;

//9. Doble punto en un nmero de tipo real. Ejemplo: 3-2..4+1  7-6.46.1+2
function TEvaluar.DoblePuntoNumero(expr: string): boolean;
var
  totalpuntos, pos: integer;
  car1: char;
begin
  totalpuntos := 0;
  for pos := 1 to length(expr) do
  begin
    car1 := expr[pos];   //Extrae un carcter
    if ((car1 < '0') or (car1 > '9')) and (car1 <> '.') then totalpuntos := 0;
    if (car1 = '.') then Inc(totalpuntos);
    if (totalpuntos > 1) then
    begin
      Result := true;
      Exit;
    end;
  end;
  Result := false;
end;

//10. Un parntesis que cierra seguido de una variable. Ejemplo: (12-4)y-1
function TEvaluar.ParentesisCierraVariable(expr: string): boolean;
var
  pos: integer;
begin
  for pos := 1 to length(expr) do
  begin
    if (expr[pos] = ')') then //Compara si el primer carcter es parntesis que cierra y el siguiente es letra
      if (expr[pos + 1] >= 'a') and (expr[pos + 1] <= 'z') then
      begin
        Result := true;
        Exit;
      end;
  end;
  Result := false;
end;

//11. Una variable seguida de un punto. Ejemplo: 4-z.1+3
function TEvaluar.VariableluegoPunto(expr: string): boolean;
var
  pos: integer;
begin
  for pos := 1 to length(expr) do
  begin
    if (expr[pos] >= 'a') and (expr[pos] <= 'z') then
      if expr[pos + 1] = '.'then
      begin
        Result := true;
        Exit;
      end;
  end;
  Result := false;
end;

//12. Un punto seguido de una variable. Ejemplo: 7-2.p+1
function TEvaluar.PuntoluegoVariable(expr: string): boolean;
var
  pos: integer;
begin
  for pos := 1 to length(expr) do
  begin
    if (expr[pos] = '.') then
      if (expr[pos + 1] >= 'a') and (expr[pos + 1] <= 'z') then
      begin
        Result := true;
        Exit;
      end;
  end;
  Result := false;
end;

//13. Un nmero antes de una variable. Ejemplo: 3x+1
//Nota: Algebraicamente es aceptable 3x+1 pero entonces vuelve ms complejo un evaluador porque debe saber que 3x+1 es en realidad 3*x+1
function TEvaluar.NumeroAntesVariable(expr: string): boolean;
var
  pos: integer;
begin
  for pos := 1 to length(expr) do
  begin
    if (expr[pos] >= '0') and (expr[pos] <= '9') then
      if (expr[pos + 1] >= 'a') and (expr[pos + 1] <= 'z') then
      begin
        Result := true;
        Exit;
      end;
  end;
  Result := false;
end;

//14. Un nmero despus de una variable. Ejemplo: x21+4
function TEvaluar.VariableDespuesNumero(expr: string): boolean;
var
  pos: integer;
begin
  for pos := 1 to length(expr) do
  begin
    if (expr[pos] >= 'a') and (expr[pos] <= 'z') then
      if (expr[pos + 1] >= '0') and (expr[pos + 1] <= '9') then
      begin
        Result := true;
        Exit;
      end;
  end;
  Result := false;
end;

//15. Chequea si hay 4 o ms variables seguidas
function TEvaluar.Chequea4letras(expr: string): boolean;
var
  pos: integer;
  car1, car2, car3, car4: char;
begin
  for pos := 1 to length(expr)-3 do
  begin
    car1 := expr[pos];
    car2 := expr[pos + 1];
    car3 := expr[pos + 2];
    car4 := expr[pos + 3];

    if (car1 >= 'a') and (car1 <= 'z') and (car2 >= 'a') and (car2 <= 'z') and (car3 >= 'a') and (car3 <= 'z') and (car4 >= 'a') and (car4 <= 'z') then
    begin
        Result := true;
        Exit;
      end;
  end;
  Result := false;
end;

//16. Si detecta tres letras seguidas y luego un parntesis que abre, entonces verifica si es funcin o no
function TEvaluar.FuncionInvalida(expr: string): boolean;
var
  pos: integer;
  car1, car2, car3: char;
begin
  for pos := 1 to length(expr)-2 do
  begin
    car1 := expr[pos];
    car2 := expr[pos + 1];
    car3 := expr[pos + 2];

    //Si encuentra tres letras seguidas
    if (car1 >= 'a') and (car1 <= 'z') and (car2 >= 'a') and (car2 <= 'z') and (car3 >= 'a') and (car3 <= 'z') then
    begin
      if pos >= length(expr) - 4 then begin Result:= true; Exit; end;//Hay un error porque no sigue parntesis
      if expr[pos + 3] <> '(' then begin Result:= true; Exit; end; //Hay un error porque no hay parntesis
      if EsFuncionInvalida(car1, car2, car3) then begin Result:= true; Exit; end;
    end;
  end;
  Result := false;
end;

//Chequea si las tres letras enviadas son una funcin
function TEvaluar.EsFuncionInvalida(car1: char; car2: char; car3: char): boolean;
var
  pos, tamfunciones:integer;
  listfunc1, listfunc2, listfunc3: char;
  listafunciones: string;
begin
  listafunciones := 'sinsencostanabsasnacsatnlogceiexpsqrrcb';
  tamfunciones := length(listafunciones);
  pos := 1;
  while (pos <= tamfunciones - 2) do
  begin
    listfunc1 := listafunciones[pos];
    listfunc2 := listafunciones[pos + 1];
    listfunc3 := listafunciones[pos + 2];

    if (car1 = listfunc1) and (car2 = listfunc2) and (car3 = listfunc3) then
    begin
      Result := false;
      exit;
    end;

    pos := pos + 3;
  end;
  Result := true;
end;

//17. Si detecta slo dos letras seguidas es un error
function TEvaluar.VariableInvalida(expr: string): boolean;
var
  cuentaletras, pos: integer;
begin
  cuentaletras := 0;
  for pos := 1 to length(expr) do
  begin
    if (expr[pos] >= 'a') and (expr[pos] <= 'z') then
      Inc(cuentaletras)
    else
    begin
      if cuentaletras = 2 then
      begin
        Result := true;
        exit;
      end;
      cuentaletras := 0;
    end;
  end;
  if cuentaletras = 2 then
    Result := true
  else
    Result := false;
end;

//18. Antes de parntesis que abre hay una letra
function TEvaluar.VariableParentesisAbre(expr: string): boolean;
var
  pos, cuentaletras: integer;
  car1: char;
begin
  cuentaletras := 0;
  for pos := 1 to length(expr) do
  begin
    car1 := expr[pos];
    if (car1 >= 'a') and (car1 <= 'z') then
      Inc(cuentaletras)
    else if (car1 = '(') and (cuentaletras = 1) then
    begin
      Result := true;
      Exit;
    end
    else
      cuentaletras := 0;
  end;
  Result := false;
end;

//19. Despus de parntesis que cierra sigue parntesis que abre. Ejemplo: (4-5)(2*x)
function TEvaluar.ParCierraParAbre(expr: string): boolean;
var
  pos: integer;
begin
      for pos:=1 to length(expr) - 1 do
      begin
          if (expr[pos]=')') and (expr[pos+1]='(') then
          begin
              Result := true;
              Exit;
          end;
      end;
      Result := false;
end;

//20. Despus de operador sigue un punto. Ejemplo: -.3+7
function TEvaluar.OperadorPunto(expr: string): boolean;
var
  pos: integer;
begin
      for pos:=1 to length(expr) - 1 do
      begin
          if (expr[pos]='+') or (expr[pos]='-') or (expr[pos]='*') or (expr[pos]='/') or (expr[pos]='^') then
                  if expr[pos+1]='.' then
                  begin
                    Result := true;
                    Exit;
                  end;
      end;
      Result := false;
end;

//21. Despus de parntesis que abre sigue un punto. Ejemplo: 3*(.5+4)
function TEvaluar.ParAbrePunto(expr: string): boolean;
var
  pos: integer;
begin
      for pos:=1 to length(expr) - 1 do
      begin
          if (expr[pos]='(') and (expr[pos+1]='.') then
          begin
              Result := true;
              Exit;
          end;
      end;
      Result := false;
end;

//22. Un punto seguido de un parntesis que abre. Ejemplo: 7+3.(2+6)
function TEvaluar.PuntoParAbre(expr: string): boolean;
var
  pos: integer;
begin
      for pos:=1 to length(expr) - 1 do
      begin
          if (expr[pos]='.') and (expr[pos+1]='(') then
          begin
              Result := true;
              Exit;
          end;
      end;
      Result := false;
end;

//23. Parntesis cierra y sigue punto. Ejemplo: (4+5).7-2
function TEvaluar.ParCierraPunto(expr: string): boolean;
var
  pos: integer;
begin
      for pos:=1 to length(expr) - 1 do
      begin
          if (expr[pos]=')') and (expr[pos+1]='.') then
          begin
              Result := true;
              Exit;
          end;
      end;
      Result := false;
end;

//24. Punto seguido de operador. Ejemplo: 5.*9+1
function TEvaluar.PuntoOperador(expr: string): boolean;
var
  pos: integer;
begin
      for pos:=1 to length(expr) - 1 do
      begin
          if (expr[pos]='.') then
            if (expr[pos+1]='+') or (expr[pos+1]='-') or (expr[pos+1]='*') or (expr[pos+1]='/') or (expr[pos+1]='^') then
            begin
                    Result := true;
                    Exit;
            end;
      end;
      Result := false;
end;

//25. Punto y sigue parntesis que cierra. Ejemplo: (3+2.)*5
function TEvaluar.PuntoParCierra(expr: string): boolean;
var
  pos: integer;
begin
      for pos:=1 to length(expr) - 1 do
      begin
          if (expr[pos]='.') and (expr[pos+1]=')') then
          begin
              Result := true;
              Exit;
          end;
      end;
      Result := false;
end;


{ Convierte una expresin con el menos unario en una expresin valida para el evaluador de expresiones:
 1. Si encuentra un - al inicio le agrega un cero
 2. Si detecta un parntesis que abre y un - entonces pone el 0 adelante
 3. Si detecta un operador y luego un menos, entonces agrega un "(0-1)#" entre esos dos
}
function TEvaluar.ArreglaNegativos(expresion: string): string;
var
  letra1: char;
  NuevaExpresion, NuevaExpresion2: string;
  pos: integer;
begin
  NuevaExpresion := '';
  NuevaExpresion2 := '';

  //Si detecta un operador y luego un menos, entonces reemplaza el menos con un "(0-1)#"
  pos := 1;
  while (pos <= length(expresion)) do
  begin
    letra1 := expresion[pos];
    if (letra1='+') or (letra1='-') or (letra1='*') or (letra1='/') or (letra1='^') then
      if expresion[pos+1]='-' then
      begin
        NuevaExpresion := NuevaExpresion + letra1 + '(0-1)#';
        Inc(pos); Inc(pos);
        continue;
      end;
    NuevaExpresion := NuevaExpresion + letra1;
    Inc(pos);
  end;

  //Si detecta un parntesis que abre y luego un menos, entonces reemplaza el menos con un "(0-1)#"
  pos := 1;
  while (pos <= length(NuevaExpresion)) do
  begin
    letra1 := NuevaExpresion[pos];
    if letra1='(' then
      if NuevaExpresion[pos+1]='-' then
      begin
        NuevaExpresion2 := NuevaExpresion2 + letra1 + '(0-1)#';
        Inc(pos); Inc(pos);
        continue;
      end;
    NuevaExpresion2 := NuevaExpresion2 + letra1;
    Inc(pos);
  end;

  Result := NuevaExpresion2;
end;



//Inicializa las listas, convierte la expresin en piezas simples y luego en piezas de ejecucin
procedure TEvaluar.Analizar(expresion: string);
begin
  PiezaSimple.Free;
  PiezaEjecuta.Free;
  PiezaSimple := TObjectList.Create;
  PiezaEjecuta := TObjectList.Create;
  Generar_Piezas_Simples(expresion);
  Generar_Piezas_Ejecucion();
end;

//Convierte la expresin en piezas simples: nmeros # parntesis # variables # operadores # funciones
procedure TEvaluar.Generar_Piezas_Simples(expresion: string);
var
  longExpresion, cont, funciondetectada, funcion: integer;
  parteentera, partedecimal, divide: double;
  entero, armanumero: boolean;
  letra, letra2, letra3: char;
begin
  longExpresion := length(expresion);

  //Variables requeridas para armar un nmero
  parteentera := 0;
  partedecimal := 0;
  divide := 1;
  entero := true;
  armanumero := false;

  cont := 1;
  while cont <= longExpresion do  //Va de letra en letra de la expresin
  begin
    letra := expresion[cont];
    if letra = '.' then  //Si letra es . entonces el resto de digitos ledos son la parte decimal del nmero
    begin
      entero := false
    end
    else if (letra >= '0') and (letra <= '9') then  //Si es un nmero, entonces lo va armando
    begin
      armanumero := true;
      if (entero) then
        parteentera := parteentera * 10 + ord(letra) - ASCIINUMERO //La parte entera del nmero
      else
      begin
        divide := divide * 10;
        partedecimal := partedecimal * 10 + ord(letra) - ASCIINUMERO; //La parte decimal del nmero
      end;
    end
    else
    begin
      if armanumero = true then //Si tena armado un nmero, entonces crea la pieza ESNUMERO
      begin
        objPiezaSimple := TPieza_Simple.Create(ESNUMERO, 0, '0', parteentera+partedecimal/divide, 0, 0);
        PiezaSimple.Add(objPiezaSimple);
        parteentera := 0;
        partedecimal := 0;
        divide := 1;
        entero := true;
        armanumero := false;
      end;
      if (letra = '+') or (letra = '-') or (letra = '*') or (letra = '/') or (letra = '^') or (letra = '#') then
      begin
        objPiezaSimple := TPieza_Simple.Create(ESOPERADOR, 0, letra, 0, 0, 0);
        PiezaSimple.Add(objPiezaSimple);
      end
      else if (letra = '(') then
      begin
        objPiezaSimple := TPieza_Simple.Create(ESPARABRE, 0, '0', 0, 0, 0);
        PiezaSimple.Add(objPiezaSimple);
      end
      else if (letra = ')') then
      begin
        objPiezaSimple := TPieza_Simple.Create(ESPARCIERRA, 0, '0', 0, 0, 0);
        PiezaSimple.Add(objPiezaSimple);
      end
      else if (letra >= 'a') and (letra <= 'z') then //Es variable o funcin?
      begin
        // Detecta si es una funcin porque tiene dos letras seguidas
        if (cont < longExpresion - 1) then
        begin
          letra2 := expresion[cont + 1]; // Chequea si el siguiente carcter es una letra, dado el caso es una funcin
          if (letra2 >= 'a') and (letra2 <= 'z') then
          begin
            letra3 := expresion[cont + 2];
            funcionDetectada := 1;  // Identifica la funcin
            funcion := 1;
            while (funcion <= TAMANOFUNCION) do
            begin
              if (letra = listaFunciones[funcion]) and (letra2 = listaFunciones[funcion + 1]) and (letra3 = listaFunciones[funcion + 2]) then break;
              Inc(funcionDetectada);
              funcion := funcion + 3;
            end;
            objPiezaSimple := TPieza_Simple.Create(ESFUNCION, funcionDetectada, '0', 0, 0, 0); //Adiciona funcin a la lista
            PiezaSimple.Add(objPiezaSimple);
            cont := cont + 3; // Mueve tres caracteres  sin(  [s][i][n][(]
          end
        else // Es una variable, no una funcin
        begin
          objPiezaSimple := TPieza_Simple.Create(ESVARIABLE, 0, '0', 0, ord(letra) - ASCIILETRA, 0);
          PiezaSimple.Add(objPiezaSimple);
        end
      end
      else // Es una variable, no una funcin
        begin
          objPiezaSimple := TPieza_Simple.Create(ESVARIABLE, 0, '0', 0, ord(letra) - ASCIILETRA, 0);
          PiezaSimple.Add(objPiezaSimple);
        end
      end;
    end;
    Inc(cont);
  end;
  if armanumero then
  begin
    objPiezaSimple := TPieza_Simple.Create(ESNUMERO, 0, '0', parteentera+partedecimal/divide, 0, 0);
    PiezaSimple.Add(objPiezaSimple);
  end
end;

//Toma las piezas simples y las convierte en piezas de ejecucin de funciones
//Acumula = funcin (operando(nmero/variable/acumula))
procedure TEvaluar.Generar_Piezas_Ejecucion();
var
  cont: integer;
begin
  cont := PiezaSimple.Count-1;
  Contador_Acumula := 0;
  repeat
    if ((PiezaSimple[cont] as TPieza_Simple).getTipo() = ESPARABRE) or ((PiezaSimple[cont] as TPieza_Simple).getTipo() = ESFUNCION) then
    begin
        Generar_Piezas_Operador('#', '#', cont);  //Primero evala los menos unarios
        Generar_Piezas_Operador('^', '^', cont);  //Luego evala las potencias
        Generar_Piezas_Operador('*', '/', cont);  //Luego evala multiplicar y dividir
        Generar_Piezas_Operador('+', '-', cont);  //Finalmente evala sumar y restar

        //Crea pieza de ejecucin
        objPiezaEjecuta := TPieza_Ejecuta.Create((PiezaSimple[cont] as TPieza_Simple).getFuncion(),
                      (PiezaSimple[cont + 1] as TPieza_Simple).getTipo(), (PiezaSimple[cont + 1] as TPieza_Simple).getNumero(), (PiezaSimple[cont + 1] as TPieza_Simple).getVariable(), (PiezaSimple[cont + 1] as TPieza_Simple).getAcumula(),
                      '+', ESNUMERO, 0, 0, 0);
        PiezaEjecuta.Add(objPiezaEjecuta);

        //La pieza pasa a ser de tipo Acumulador
        (PiezaSimple[cont + 1] as TPieza_Simple).setAcumula(Contador_Acumula);
        Inc(Contador_Acumula);

        //Quita el parntesis/funcin que abre y el que cierra, dejando el centro
        PiezaSimple.Delete(cont);
        PiezaSimple.Delete(cont + 1);
    end;
    cont := cont - 1;
  until cont < 0;
end;

//Toma las piezas simples y las convierte en piezas de ejecucin
//Acumula = operando(nmero/variable/acumula)  operador(+, -, *, /, ^)   operando(nmero/variable/acumula)
procedure TEvaluar.Generar_Piezas_Operador(operA: char; operB: char; inicio: integer);
var
  cont: integer;
begin
  cont := inicio + 1;
  repeat
    if ((PiezaSimple[cont] as TPieza_Simple).getTipo() = ESOPERADOR) AND ( ((PiezaSimple[cont] as TPieza_Simple).getOperador() = operA) OR ((PiezaSimple[cont] as TPieza_Simple).getOperador() = operB) ) then
    begin
      //Crea pieza de ejecucin
      objPiezaEjecuta := TPieza_Ejecuta.Create(0,
                      (PiezaSimple[cont - 1] as TPieza_Simple).getTipo(),
                      (PiezaSimple[cont - 1] as TPieza_Simple).getNumero(), (PiezaSimple[cont - 1] as TPieza_Simple).getVariable(), (PiezaSimple[cont - 1] as TPieza_Simple).getAcumula(),
                      (PiezaSimple[cont] as TPieza_Simple).getOperador(),
                      (PiezaSimple[cont + 1] as TPieza_Simple).getTipo(),
                      (PiezaSimple[cont + 1] as TPieza_Simple).getNumero(), (PiezaSimple[cont + 1] as TPieza_Simple).getVariable(), (PiezaSimple[cont + 1] as TPieza_Simple).getAcumula());
      PiezaEjecuta.Add(objPiezaEjecuta);

      //Elimina la pieza del operador y la siguiente
      PiezaSimple.Delete(cont);
      PiezaSimple.Delete(cont);

      //Retorna el contador en uno para tomar la siguiente operacin
      cont := cont - 1;

      //Cambia la pieza anterior por pieza acumula
      (PiezaSimple[cont] as TPieza_Simple).setAcumula(Contador_Acumula);
      Inc(Contador_Acumula);
    end;
    Inc(cont);
  until (cont >= PiezaSimple.Count) or ((PiezaSimple[cont] as TPieza_Simple).getTipo() = ESPARCIERRA);
end;

//Calcula la expresin convertida en piezas de ejecucin
function TEvaluar.Calcular(): double;
var
  valorA, valorB: double;
  totalPiezaEjecuta, cont: integer;
begin
  valorA := 0;
  valorB := 0;
  totalPiezaEjecuta := PiezaEjecuta.Count;

  try
    for cont := 0 to totalPiezaEjecuta-1 do
    begin
      case (PiezaEjecuta[cont] as TPieza_Ejecuta).getTipoOperA() of
        5: begin valorA := (PiezaEjecuta[cont] as TPieza_Ejecuta).getNumeroA(); end; //Es un nmero?
        6: begin valorA := VariableAlgebra[(PiezaEjecuta[cont] as TPieza_Ejecuta).getVariableA()]; end;  //Es una variable?
        7: begin valorA := (PiezaEjecuta[(PiezaEjecuta[cont] as TPieza_Ejecuta).getAcumulaA()] as TPieza_Ejecuta).getValorPieza(); end; //Es una expresin anterior?
      end;

      case (PiezaEjecuta[cont] as TPieza_Ejecuta).getFuncion() of
        0:begin
            if (PiezaEjecuta[cont] as TPieza_Ejecuta).getTipoOperB() = 5 then valorB := (PiezaEjecuta[cont] as TPieza_Ejecuta).getNumeroB() //Es un nmero?
            else if (PiezaEjecuta[cont] as TPieza_Ejecuta).getTipoOperB() = 6 then valorB := VariableAlgebra[(PiezaEjecuta[cont] as TPieza_Ejecuta).getVariableB()] //Es una variable?
            else if (PiezaEjecuta[cont] as TPieza_Ejecuta).getTipoOperB() = 7 then valorB := (PiezaEjecuta[(PiezaEjecuta[cont] as TPieza_Ejecuta).getAcumulaB()] as TPieza_Ejecuta).getValorPieza(); //Es una expresin anterior?

            if (PiezaEjecuta[cont] as TPieza_Ejecuta).getOperador() = '#' then (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(valorA * valorB)
            else if (PiezaEjecuta[cont] as TPieza_Ejecuta).getOperador() = '+' then (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(valorA + valorB)
            else if (PiezaEjecuta[cont] as TPieza_Ejecuta).getOperador() = '-' then (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(valorA - valorB)
            else if (PiezaEjecuta[cont] as TPieza_Ejecuta).getOperador() = '*' then (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(valorA * valorB)
            else if (PiezaEjecuta[cont] as TPieza_Ejecuta).getOperador() = '/' then (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(valorA / valorB)
            else if (PiezaEjecuta[cont] as TPieza_Ejecuta).getOperador() = '^' then (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(power(valorA, valorB));
          end;
        1, 2: begin (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(sin(valorA)); end;
        3: begin (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(cos(valorA)); end;
        4: begin (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(tan(valorA)); end;
        5: begin (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(abs(valorA)); end;
        6: begin (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(arcsin(valorA)); end;
        7: begin (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(arccos(valorA)); end;
        8: begin (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(arctan(valorA)); end;
        9: begin (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(ln(valorA)); end;
        10: begin (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(ceil(valorA)); end;
        11: begin (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(exp(valorA)); end;
        12: begin (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(sqrt(valorA)); end;
        13: begin (PiezaEjecuta[cont] as TPieza_Ejecuta).setValorPieza(power(valorA, 0.333333333333)); end;
      end;
    end;
  except //Captura el error matemtico
      on EMathError do
      begin
        Result := NaN;
        Exit;
      end;
  end;
  Calcular := (PiezaEjecuta[totalPiezaEjecuta - 1] as TPieza_Ejecuta).getValorPieza();
end;

// Da valor a las variables que tendr la expresin algebraica
procedure TEvaluar.ValorVariable(variableAlgebra: char; valor: double);
begin
  self.VariableAlgebra[ord(variableAlgebra) - self.ASCIILETRA] := valor;
end;


end.
