/*  Autor: Rafael Alberto Moreno Parra
  Sitio Web:  http://darwin.50webs.com
  Correo:  enginelife@hotmail.com

  Evaluador de expresiones algebraicas por ejemplo, 57+1.87/84.89-(6.8*e+b+8769-4*b/8+b^2*4^e/f)+5.4-(d/9.6+0.2) con las siguientes funcionalidades:
  1. Ilimitado nmero de parntesis
  2. Ms rpido y ms sencillo que el evaluador escrito para el primer libro: "Desarrollo de un evaluador de expresiones algebraicas"
  3. Manejo de 26 variables
  4. Manejo de 12 funciones
  5. Manejo de operadores +, -, *, /, ^ (potencia)
  6. Manejo del menos unario: lo reemplaza con (0-1)# donde # es la operacin con mayor relevancia y equivale a la multiplicacin

  Versin: 2.00 [Genera el libro "Desarrollo de un evaluador de expresiones algebraicas. Versin 2.0"]
  Fecha: Enero de 2013
  Licencia: LGPL

  Algoritmo:

  Se toma una expresin como 7*x+sen(12.8/y+9)-5*cos(9-(8.3/5.11^3-4.7)*7.12)+0.445
  Se agregan parntesis de inicio y fin  (7*x+sen(12.8/y+9)-5*cos(9-(8.3/5.11^3-4.7)*7.12)+0.445)
  Se divide en piezas simples   | ( | 7 | * | x | + | sen( | 12.8 | / | y | + | 9 | ) | - | 5 | * | cos( | 9 | - | ( | 8.3 | / | 5.11 | ^ | 3 | - | 4.7 | ) | * | 7.12 | ) | + | 0.445 | ) |
  Esas piezas estn clasificadas:
   Parntesis que abre (
   Parntesis que cierra )
   Nmeros  7   12.8  9   5
   Variables  x  y
   Operadores + - * / ^
   Funciones  sen(  cos(
  Luego se convierte esa expresin larga en expresiones cortas de ejecucin del tipo
  Acumula = operando(nmero/variable/acumula)  operador(+, -, *, /, ^)   operando(nmero/variable/acumula)
    [0]  5.11 ^ 3
    [1]  8.3  / [0]
    [2]  [1] - 4.7
    [3]  [2] + 0
    [4]  [3] * 7,12
    [5]  9   - [4]
    [6]  cos([5])
    [7]  12,8 / y
    [8]  [7] +  9
    [9]  sen([8])
    [10] 7 * x
    [11] 5 * [6]
    [12] [10] + [9]
    [13] [12] - [11]
    [14] [13] + 0.445
    [15] [14] + 0
  La expresin ya est analizada y lista para evaluar.
  Se evala yendo de [0] a [15], en [15] est el valor final.
*/

#include "Pieza_Simple.h"
#include "Pieza_Ejecuta.h"
#include <vector>

class Evaluar
{
private:
    /* Esta constante sirve para que se reste al carcter y se obtenga el nmero.  Ejemplo:  '7' - ASCIINUMERO =  7 */
    static const int ASCIINUMERO = 48;

    /* Esta constante sirve para que se reste al carcter y se obtenga el nmero de la letra. Ejemplo:  'b' - ASCIILETRA =  1 */
    static const int ASCIILETRA = 97;

    /* Las funciones que soporta este evaluador */
    static const int TAMANOFUNCION = 39;
    static const char *listaFunciones;

    /* Constantes de los diferentes tipos de datos que tendrn las piezas */
    static const int ESFUNCION = 1;
    static const int ESPARABRE = 2;
    static const int ESPARCIERRA = 3;
    static const int ESOPERADOR = 4;
    static const int ESNUMERO = 5;
    static const int ESVARIABLE = 6;
    static const int ESACUMULA = 7;

    //Listado de Piezas de anlisis
    std::vector<Pieza_Simple> PiezaSimple;

    //Listado de Piezas de ejecucin
    std::vector<Pieza_Ejecuta> PiezaEjecuta;
    int Contador_Acumula;

    //Almacena los valores de las 26 diferentes variables que puede tener la expresin algebraica
    double VariableAlgebra[26];

    char *expr;
    bool DobleTripleOperadorSeguido(char *);
    bool OperadorParentesisCierra(char *);
    bool ParentesisAbreOperador(char *);
    bool ParentesisDesbalanceados(char *);
    bool ParentesisVacio(char *);
    bool ParentesisBalanceIncorrecto(char *);
    bool ParentesisCierraNumero(char *);
    bool NumeroParentesisAbre(char *);
    bool DoblePuntoNumero(char *);
    bool ParentesisCierraVariable(char *);
    bool VariableluegoPunto(char *);
    bool PuntoluegoVariable(char *);
    bool NumeroAntesVariable(char *);
    bool VariableDespuesNumero(char *);
    bool Chequea4letras(char *);
    bool FuncionInvalida(char *);
    bool EsFuncionInvalida(char, char, char);
    bool VariableInvalida(char *);
    bool VariableParentesisAbre(char *);
    bool ParCierraParAbre(char *);
    bool OperadorPunto(char *);
    bool ParAbrePunto(char *);
    bool PuntoParAbre(char *);
    bool ParCierraPunto(char *);
    bool PuntoOperador(char *);
    bool PuntoParCierra(char *);

    //Convierte la expresin en piezas simples: nmeros # parntesis # variables # operadores # funciones
    void Generar_Piezas_Simples(char *expresion);

    //Toma las piezas simples y las convierte en piezas de ejecucin de funciones
    //Acumula = funcin (operando(nmero/variable/acumula))
    void Generar_Piezas_Ejecucion();

    //Toma las piezas simples y las convierte en piezas de ejecucin
    //Acumula = operando(nmero/variable/acumula)  operador(+, -, *, /, ^)   operando(nmero/variable/acumula)
    void Generar_Piezas_Operador(char operA, char operB, int inicio);



public:
    int EvaluaSintaxis(char *);
    void MensajeSintaxis(int CodigoError, char *);

    void TransformaExpresion(char *, char *);
    int ArreglaNegativos(char *, char *);

    //Inicializa las listas, convierte la expresin en piezas simples y luego en piezas de ejecucin
    void Analizar(char *expresion);

    //Calcula la expresin convertida en piezas de ejecucin
    double Calcular();

    // Da valor a las variables que tendr la expresin algebraica
    void ValorVariable(char variableAlgebra, double valor);
};